from dataclasses import dataclass, field

from LMS.Common.LMS_DataType import LMS_DataType
from LMS.TitleConfig.Definitions.Value import ValueDefinition


class LMS_Field:
    """A class that represents a mapped field linked to a config definition.

    Acts as values for Attributes and Tag Parameters."""

    def __init__(
        self, value: int | str | float | bytes | bool, definition: ValueDefinition
    ):
        self._definition = definition
        self._verify_value(value)
        self._value = value

    def __repr__(self):
        typename = self._definition.datatype.name
        if typename == "LIST":
            if len(self.list_items) > 6:
                preview = self.list_items[:3] + ["..."]
            else:
                preview = self.list_items
            return f"LMS_Field(value={self._value}, options={preview})"
        return f"LMS_Field(value={self._value!r}, type={typename})"

    @property
    def name(self) -> str:
        """The name of the field."""
        return self._definition.name

    @property
    def description(self) -> str:
        """The description of the field."""
        return self._definition.description

    @property
    def value(self) -> int | str | float | bytes | bool:
        """The value of the field instance."""
        return self._value

    @property
    def datatype(self) -> LMS_DataType:
        """The datatype of the field instance."""
        return self._definition.datatype

    @property
    def list_items(self) -> list[str]:
        """The list items bound to the field instance. Only is valid for LMS_Datatype.LIST values."""
        return self._definition.list_items

    @value.setter
    def value(self, new_value: int | str | float | bytes | bool):
        self._verify_value(new_value)
        self._value = new_value

    def _verify_value(self, value: int | str | float | bytes | bool) -> None:
        datatype = self._definition.datatype

        # Check if the value is not the correct instance of the datatype
        if type(value) is not (builtin_type := datatype.builtin_type):
            raise TypeError(
                f"The value provided for '{self.name}' type '{type(value)}' should be '{builtin_type.__name__}'."
            )

        # These values require no extra verification
        if datatype in (LMS_DataType.BOOL, LMS_DataType.STRING):
            return

        match datatype:
            case LMS_DataType.BYTE:
                if len(value) != 1:
                    raise ValueError("Byte types only work for values of length 1!")
                else:
                    return
            case LMS_DataType.LIST:
                if value not in self._definition.list_items:
                    raise ValueError(
                        f"The value of '{value}' provided for field '{self.name}' is not in the list {self._definition.list_items}."
                    )
                else:
                    return
            case LMS_DataType.FLOAT32:
                min_value, max_value = -3.4028235e38, 3.4028235e38

            # Other number types fall here
            case _:
                bits = datatype.stream_size * 8
                if datatype.signed:
                    max_value = 2 ** (bits - 1)
                    min_value = -max_value
                else:
                    min_value, max_value = 0, (2**bits) - 1

        if not min_value <= value <= max_value:
            raise ValueError(
                f"The value '{value}' of type '{datatype}' provided for field '{self.name}' is out of range of ({min_value}, {max_value})"
            )


def cast_value(value: int | str | float | bytes | bool, datatype: LMS_DataType) -> str:
    if datatype in (LMS_DataType.STRING, LMS_DataType.LIST):
        return value

    match datatype:
        case LMS_DataType.BYTE:
            return hex(int(value))
        case LMS_DataType.BOOL:
            return value.strip().lower() == "true"
        case LMS_DataType.FLOAT32:
            try:
                return float(value)
            except ValueError:
                raise ValueError(f"Cannot convert '{value}' to float.")
        case _:
            if value.isdigit():
                return int(value)
            else:
                raise TypeError(
                    f"The provided value for type '{datatype}' is not a digit!"
                )



from dataclasses import dataclass, field

from LMS.TitleConfig.Definitions.Value import ValueDefinition


@dataclass(frozen=True)
class AttributeConfig:
    """Class that represents an attribute config definition"""
    name: str
    description: str = field(repr=False)
    definitions: list[ValueDefinition]



from importlib import resources
from typing import Self

import yaml

from LMS.Common.LMS_DataType import LMS_DataType
from LMS.Project.MSBP import MSBP
from LMS.TitleConfig.Definitions.Attributes import AttributeConfig
from LMS.TitleConfig.Definitions.Tags import TagConfig, TagDefinition
from LMS.TitleConfig.Definitions.Value import ValueDefinition

TAG_KEY = "tag_definitions"
ATTR_KEY = "attribute_definitions"


class TitleConfig:
    """Represents a configuration for a specific game/title."""

    # Populate the preset list from the directory
    PRESET_LIST = [
        file.name.removesuffix(".yaml")
        for file in resources.files("LMS.TitleConfig.Presets").iterdir()
        if file.name.endswith(".yaml")
    ]

    def __init__(
        self, attribute_configs: dict[str, AttributeConfig], tag_config: TagConfig
    ):
        self._attribute_configs = attribute_configs
        self._tag_config = tag_config

    @property
    def attribute_configs(self) -> dict[str, AttributeConfig]:
        """The map of attribute config instances."""
        return self._attribute_configs

    @property
    def tag_config(self) -> TagConfig:
        """The loaded tag config instance."""
        return self._tag_config

    @classmethod
    def load_preset(cls, game: str) -> Self:
        """Loads an existing preset from a game."""
        if game.lower() not in [preset.lower() for preset in cls.PRESET_LIST]:
            raise FileNotFoundError(f"Preset '{game}' not found.")

        with resources.open_text("LMS.TitleConfig.Presets", f"{game}.yaml") as f:
            return cls.load_config(f.read())

    @classmethod
    def load_file(self, file_path: str) -> Self:
        """Loads a config from a file.

        :param file_path: the path to the config."""
        with open(file_path, "r") as f:
            return TitleConfig.load_config(f.read())

    @classmethod
    def load_config(cls, content: str | dict) -> Self:
        """Loads the config of a specified game.

        :param content: the config content, as a string or loaded as a dictionary."""
        if isinstance(content, str):
            parsed_content = yaml.safe_load(content)
        else:
            parsed_content = content

        # Load the attribute definitions
        attribute_configs = {}
        for config in parsed_content[ATTR_KEY]:
            definitions = [
                ValueDefinition.from_dict(value_def)
                for value_def in config["definitions"]
            ]
            attribute_configs[config["name"]] = AttributeConfig(
                config["name"], config["description"], definitions
            )

        # Load the tag definitions
        tag_definitions = []
        group_map = parsed_content[TAG_KEY]["groups"]
        for tag_def in parsed_content[TAG_KEY]["tags"]:
            tag_definitions.append(TagDefinition.from_dict(tag_def, group_map))

        tag_config = TagConfig(group_map, tag_definitions)

        return cls(attribute_configs, tag_config)

    @staticmethod
    def generate_file(file_path: str, project: MSBP) -> None:
        with open(file_path, "w+") as f:
            yaml.safe_dump(
                TitleConfig.generate_config(project),
                f,
                default_flow_style=False,
                sort_keys=False,
            )

    @staticmethod
    def generate_config(project: MSBP) -> dict | None:
        """Creates a message config file for the game.

        :param project: a MSBP object."""
        # TODO: Add custom node definitions

        config = {}
        # Source files may be a path to a non-existent directory.
        # These files were generated from the source machine from the actual libMS tool
        # Shorten the filename with basname and replace the extension with .msbt for lookup later when reading a MSBT
        config[TAG_KEY] = {
            "groups": {
                i + 1: group.name for i, group in enumerate(project.tag_groups[1:])
            },
            "tags": [],
        }

        # Slice to exclude System group
        if project.tag_groups is not None:
            for group_i, group in enumerate(project.tag_groups[1:], start=1):
                for tag_i, info in enumerate(group.tag_definitions):

                    definition = {
                        "name": info.name,
                        "group_index": group_i,
                        "tag_index": tag_i,
                        "description": "",
                        "parameters": [],
                    }

                    for param_info in info.param_info:
                        param_definition = {
                            "name": param_info.name,
                            "description": "",
                            "datatype": param_info.datatype.to_string(),
                        }

                        if param_info.datatype is LMS_DataType.LIST:
                            param_definition["list_items"] = param_info.list_items

                        definition["parameters"].append(param_definition)

                    config[TAG_KEY]["tags"].append(definition)

        # Set main attribute entries as the definitions from the MSBP
        # Since most games use one MSBP these act the primary definitions
        config[ATTR_KEY] = []

        if project.attribute_info is not None:

            attr_definitions = []
            for attr_info in project.attribute_info:
                definition = {
                    "name": attr_info.name,
                    "description": "",
                    "datatype": attr_info.datatype.to_string(),
                }
                if attr_info.datatype is LMS_DataType.LIST:
                    definition["list_items"] = attr_info.list_items

                attr_definitions.append(definition)

            # Main attribute entries
            config[ATTR_KEY].append(
                {
                    "name": project.name,
                    "description": "",
                    "definitions": attr_definitions,
                }
            )

        return config
